import copy
import os

import pytest

from poglink.bot import BotConfig, ConfigurableBot
from poglink.error import ConfigReadError


@pytest.fixture
def sample_config_dict():
    return {
        "token": "abcd",
        "rates_channel_id": "1234",
        "bans_channel_id": "4321",
        "polling_delay": 69,
        "allowed_roles": ["admin"],
        "rates_urls": ["www.arkrates.com"],
        "data_dir": "/my/fake/dir",
        "send_embed_on_startup": True,
    }


@pytest.fixture
def sample_config_dict_singular(sample_config_dict):
    config = copy.deepcopy(sample_config_dict)
    config.update(
        {"allowed_roles": "regular_user", "rates_urls": {"url": "www.google.com"}}
    )
    return config


def test_botconfig_from_dict(sample_config_dict):
    botconfig = BotConfig.from_dict(sample_config_dict)

    assert botconfig.token == "abcd"
    assert botconfig.rates_channel_id == "1234"
    assert botconfig.bans_channel_id == "4321"
    assert botconfig.polling_delay == 69
    assert botconfig.allowed_roles == ["admin"]
    assert botconfig.rates_urls == ["www.arkrates.com"]
    assert botconfig.data_dir == "/my/fake/dir"


def test_botconfig_singular_vals(sample_config_dict_singular):
    botconfig = BotConfig.from_dict(sample_config_dict_singular)

    assert botconfig.allowed_roles == ["regular_user"]
    assert botconfig.rates_urls == {"url": "www.google.com"}


def test_botconfig_from_file(
    sample_application_config_yaml, sample_application_config_json, config_yaml
):
    botconfig_yaml = BotConfig.from_file(sample_application_config_yaml)
    botconfig_json = BotConfig.from_file(sample_application_config_json)

    # YAML and JSON formats both equivalent
    assert botconfig_yaml == botconfig_json

    # Expected values are present
    assert botconfig_yaml.token == "abcd"
    assert botconfig_yaml.rates_channel_id == "1234"
    assert botconfig_yaml.polling_delay == 60
    assert botconfig_yaml.allowed_roles == ["admin", "regular_users"]
    assert botconfig_yaml.rates_urls == [
        {"url": "http://arkdedicated.com/dynamicconfig.ini"}
    ]
    assert botconfig_yaml.data_dir is None
    assert botconfig_yaml.send_embed_on_startup is True

    # Raises error when file doesn't exist
    with pytest.raises(FileNotFoundError):
        BotConfig.from_file("/nonsense/path")

    # Grabs file called config.yaml if provided path is a directory
    botconfig = BotConfig.from_file(os.path.dirname(config_yaml))
    assert botconfig == botconfig_yaml


def test_configurable_bot(sample_config_dict):
    bot = ConfigurableBot(command_prefix=".test", config_dict=sample_config_dict)

    assert bot.config.token == "abcd"
    assert bot.config.rates_channel_id == "1234"
    assert bot.config.bans_channel_id == "4321"
    assert bot.config.polling_delay == 69
    assert bot.config.allowed_roles == ["admin"]
    assert bot.config.rates_urls == ["www.arkrates.com"]
    assert bot.config.data_dir == "/my/fake/dir"
    assert bot.config.send_embed_on_startup is True


def test_comma_separation(sample_application_config_comma_yaml):
    botconfig = BotConfig.from_file(sample_application_config_comma_yaml)

    assert botconfig.allowed_roles == ["admin", "regular_users"]
    assert botconfig.rates_urls == [
        {"url": "http://arkdedicated.com/dynamicconfig.ini"},
        {"url": "http://www.google.com"},
    ]


def test_config_read_error(application_config_broken):
    with pytest.raises(ConfigReadError):
        BotConfig.from_file(application_config_broken)



import asyncio
import logging
import os
import time

import aiohttp
import discord
from discord.ext import commands

from poglink.error import RatesFetchError, RatesProcessError
from poglink.models import RatesStatus

logger = logging.getLogger(__name__)

EMBED_IMAGE = "https://i.stack.imgur.com/Fzh0w.png"
RATE_LIMIT_DELAY = 1
DEFAULT_COLOR = "0xF8DE74"
DEFAULT_SERVER_NAME = "Unknown Server Name"
DEFAULT_URL = "Unknown URL"

# create cog class
class Rates(commands.Cog):
    def __init__(self, client):
        self.client = client

        self.rates_dicts = client.config.rates_urls
        self.channel_id = client.config.rates_channel_id
        self.polling_delay = client.config.polling_delay
        self.allowed_roles = client.config.allowed_roles
        self.data_dir = os.path.expanduser(client.config.data_dir)
        self.last_rates = [None for _ in self.rates_dicts]
        self.stable_rates = [None for _ in self.rates_dicts]
        self.consecutive_count = [0 for _ in self.rates_dicts]
        # Create parent directory for persistent data if it doesn't exist yet
        if not os.path.exists(self.data_dir):
            logger.info(f"Data directory doesn't exist yet; creating: {self.data_dir}")
            os.makedirs(self.data_dir)
        self.send_embed_on_startup = client.config.send_embed_on_startup
        self.rate_limit_delay = RATE_LIMIT_DELAY  # Can be overridden manually, but not part of the config when instantiated

    @staticmethod
    async def get_current_rates(url):
        logger.debug(f"Requesting current server rates from {url}")
        async with aiohttp.ClientSession() as session:
            try:
                async with session.get(
                    url,
                    headers={"Pragma": "no-cache", "Cache-Control": "no-cache"},
                ) as response:
                    response = await response.text()
                    rates = RatesStatus.from_raw(response)
                    logger.debug(f"Obtained server rates: {RatesStatus}")
                    return rates
            except ValueError as e:
                raise RatesProcessError(e) from e
            except Exception as e:
                raise RatesFetchError(e) from e

    async def send_embed(self, description, rates_dict, **kwargs):
        url = rates_dict.get("url", DEFAULT_URL)
        server_name = rates_dict.get("server_name", DEFAULT_SERVER_NAME)
        color = rates_dict.get("color", DEFAULT_COLOR)
        if isinstance(color, str):
            server_color = int(color, 16)
        else:
            server_color = int(color)

        # generate embed
        logger.debug(f"Attempting to send embed. desc: {description}, url: {url}")

        # generate dynamic timestamp (https://hammertime.djdavid98.art/)
        ts = int(time.time() // 60 * 60)
        ts_string = f"<t:{ts}:t>"
        title = f"{server_name} server rates updated at {ts_string}"

        embed = discord.Embed(
            description=description,
            title=title,
            color=server_color,
        )
        embed.set_image(url=EMBED_IMAGE)

        # send embed
        channel = self.client.get_channel(self.channel_id)
        message = await channel.send(embed=embed)

        # if in announcement channel, publish message
        if message.channel.type == discord.ChannelType.news:
            logger.info("Announcement channel detected: Publishing message")
            await message.publish()

    async def compare_and_notify_all(self, **kwargs):

        for idx in range(len(self.rates_dicts)):
            url = self.rates_dicts[idx].get("url")

            # Fetch current rates from online
            try:
                current_rates = await self.get_current_rates(url)
            except RatesFetchError as e:
                logger.error(f"Could not retrieve rates from ARK Web API at {url}: {e}")
                continue
            except RatesProcessError as e:
                logger.error(f"Could not process rates URL {url}: {e}")
                continue
            except Exception as e:  # pragma: no cover
                logger.error(e)
                continue

            # Only if last rates exist, get diff
            if self.last_rates[idx]:
                rates_diff = self.last_rates[idx].get_diff(current_rates)

                # If there is a difference, reset consecutive count; otherwise increment it
                if rates_diff.items:
                    self.consecutive_count[idx] = 1
                    self.last_rates[idx] = current_rates
                    logger.info(
                        f"Rates for {url} changed. Resetting consecutive rates count to 1"
                    )
                else:
                    self.consecutive_count[idx] += 1
                    logger.info(
                        f"Rates for {url} unchanged. Consecutive count = {self.consecutive_count[idx]}"
                    )

                if self.consecutive_count[idx] == 2:
                    logger.info(
                        "New rates have become stable. Comparing against previous stable rates."
                    )
                    if self.stable_rates[idx]:
                        stable_diff = self.stable_rates[idx].get_diff(current_rates)
                        if stable_diff.items:
                            # generate and send embed
                            logger.info(
                                f"Rates at {url} changed since last stable value - sending embed"
                            )
                            embed_description = stable_diff.to_embed()
                            await self.send_embed(
                                embed_description, rates_dict=self.rates_dicts[idx]
                            )
                    else:
                        logger.info(
                            f"No previous stable rates recorded at {url}. Updating new stable value, but no updates to send."
                        )

                    self.stable_rates[idx] = current_rates
            else:
                self.consecutive_count[idx] = 1
                logger.info(f"No previous rates from {url} stored yet; skipping.")

            # Update last rates value for next iteration
            self.last_rates[idx] = current_rates
            await asyncio.sleep(self.rate_limit_delay)

    # Events
    @commands.Cog.listener()
    async def on_ready(self):  # pragma: no cover
        logger.info("Cog Ready: Rates")

        # send embed with initial rates for each url upon startup.
        # TODO: Add test for this
        if self.send_embed_on_startup:
            for rates_dict in self.rates_dicts:
                rates = await self.get_current_rates(rates_dict.get("url"))
                await self.send_embed(rates.to_embed(), rates_dict)
                await asyncio.sleep(self.rate_limit_delay)

        while True:
            await self.compare_and_notify_all()
            await asyncio.sleep(self.polling_delay)


# add cog to client
def setup(client):  # pragma: no cover
    client.add_cog(Rates(client))



import logging
import os

import yaml

from poglink.error import ConfigReadError
from poglink.utils import parse_bool, parse_list

logger = logging.getLogger(__name__)

MIN_POLLING_DELAY = 5

DEFAULT_CONFIG = {
    "allowed_roles": [],
    "polling_delay": 60,
    "rates_urls": [
        {
            "server_name": "Official",
            "url": "https://cdn2.arkdedicated.com/asa/dynamicconfig.ini",
            "color": "0x63BCC3",
        }
    ],
    # "bans_url": "http://arkdedicated.com/bansummary.txt",
    "rates_channel_id": None,
    # "bans_channel_id": None, # TODO: Reimplement when bans are enabled
    "token": None,
    "data_dir": "~/.poglink",
    "send_embed_on_startup": False,
}

REQUIRED_VALUES = [
    "token",
    "rates_channel_id",
    # "bans_channel_id", # TODO: Reimplement when bans are enabled
]

LIST_VALUES = [
    "allowed_roles",
]

BOOLEAN_VALUES = ["send_embed_on_startup"]


def setup_config(args, default_config=DEFAULT_CONFIG):
    # Attempt to load config values from file if provided
    data_dir = os.path.expanduser(
        args.data_dir or os.getenv("BOT_DATA_DIR") or DEFAULT_CONFIG.get("data_dir")
    )
    config_path = os.path.join(data_dir, "config.yaml")
    logger.debug(f"Setting up configuration. Checking for config file: {config_path}")

    if os.path.exists(config_path):
        try:
            with open(os.path.expanduser(config_path)) as f:
                config_from_file = yaml.safe_load(f)
        except Exception as e:
            logger.error(f"Problem reading configuration file at {config_path}: {e}")
            raise ConfigReadError(e) from e
    else:
        logger.warning(
            f"No configuration file found at {config_path}. Configuration must be set via CLI args or environment variables."
        )
        config_from_file = {}

    # For each configuration value, attempt to obtain value in the specified order of priority:
    config = {}
    for key, default_val in default_config.items():
        config[key] = (
            getattr(args, key)
            or config_from_file.get(key)
            or os.getenv(f"BOT_{key.upper()}")
            or default_val
        )

    # handle special cases for boolean values
    for param in BOOLEAN_VALUES:
        if not isinstance(config[param], bool):
            config[param] = parse_bool(config[param])

    # handle special cases for list parsing
    for val in LIST_VALUES:
        if isinstance(config[val], str):
            config[val] = parse_list(config[val])
        elif isinstance(config[val], list):
            pass
        else:
            logger.warning(
                f"Incorrect variable format for {val}; should be comma separated list. Actual value: {config[val]}"
            )

    # handle special case for polling delay
    if config["polling_delay"] < MIN_POLLING_DELAY:
        logger.warning(
            f"Polling delay ({config['polling_delay']}) below minimum value. Setting to minimum value ({MIN_POLLING_DELAY})."
        )
        config["polling_delay"] = MIN_POLLING_DELAY

    logger.debug(f"Configuration loaded from sources: {config}")
    return config



import argparse
import logging
import os
import tempfile

import pytest
import yaml

from poglink.config import DEFAULT_CONFIG, setup_config
from poglink.error import ConfigReadError


@pytest.fixture
def cli_config_vals():
    vals = {k: None for k in DEFAULT_CONFIG.keys()}
    vals.update(
        {
            "bans_channel_id": "54321",
            "rates_channel_id": "2468",
            "polling_delay": 2,
        }
    )
    return vals


@pytest.fixture
def args(cli_config_vals):
    return argparse.Namespace(**cli_config_vals)


@pytest.fixture
def file_config_vals():
    return {
        "token": "abcdef",
        "rates_channel_id": "12345",
        "allowed_roles": ["test", "admin"],
    }


@pytest.fixture
def config_dir(file_config_vals):
    with tempfile.TemporaryDirectory() as dirname:
        fullpath = os.path.join(dirname, "config.yaml")
        with open(fullpath, "w+") as f:
            yaml.safe_dump(file_config_vals, f)

        yield dirname


@pytest.fixture
def bad_config_dir(file_config_vals):
    with tempfile.TemporaryDirectory() as dirname:
        fullpath = os.path.join(dirname, "config.yaml")
        with open(fullpath, "w+") as f:
            f.write("this\nis: a bad\t\n\t config file::::")

        yield dirname


@pytest.fixture
def env(config_dir):
    os.environ["BOT_DATA_DIR"] = config_dir
    os.environ["BOT_TOKEN"] = "fedcba"
    os.environ["BOT_SEND_EMBED_ON_STARTUP"] = "1"


def test_setup_config(args, env, caplog):
    config = setup_config(args)

    # Values are read from CLI
    assert (
        config.get("bans_channel_id") is None
    )  # "54321", # TODO: Reimplement when bans are enabled

    # Values are read from File
    assert config.get("allowed_roles") == ["test", "admin"]

    # Default value is populated
    assert (
        config.get("bans_url") is None
    )  # "http://arkdedicated.com/bansummary.txt", # TODO: Reimplement when bans are enabled

    # File takes precedence over env
    assert config.get("token") == "abcdef"

    # CLI takes precedence over file
    assert config.get("rates_channel_id") == "2468"

    # Min polling delay is enforced and logged
    assert config.get("polling_delay") == 5
    with caplog.at_level(logging.WARNING):
        assert "below minimum value" in caplog.text

    # boolean values are interpreted from env var
    assert config.get("send_embed_on_startup") is True


def test_setup_config_bad_config_file(args, caplog, bad_config_dir):

    # Test bad config file
    args.data_dir = bad_config_dir
    with pytest.raises(ConfigReadError):
        setup_config(args)
    with caplog.at_level(logging.ERROR):
        assert "Problem reading configuration file" in caplog.text

    # Test nonexistent config file
    args.data_dir = "/dir/that/does/not/exist"
    setup_config(args)
    with caplog.at_level(logging.WARNING):
        assert "No configuration file found at" in caplog.text


def test_setup_config_bad_list(env, args, caplog):

    setup_config(args)

    args.allowed_roles = 1234
    setup_config(args)
    with caplog.at_level(logging.WARNING):
        assert "Incorrect variable format" in caplog.text
